from casadi import *

# Degree of interpolating polynomial
d = 3

# Get collocation points
tau_root = collocationPoints(d, 'legendre')

# Coefficients of the collocation equation
C = numpy.zeros((d+1,d+1))

# Coefficients of the continuity equation
D = numpy.zeros(d+1)

# Coefficients of the quadrature function
B = numpy.zeros(d+1)

# Construct polynomial basis
for j in range(d+1):
    # Construct Lagrange polynomials to get the polynomial basis at the collocation point
    p = numpy.poly1d([1])
    for r in range(d+1):
        if r != j:
            p *= numpy.poly1d([1, -tau_root[r]]) / (tau_root[j]-tau_root[r])
                
    # Evaluate the polynomial at the final time to get the coefficients of the continuity equation
    D[j] = p(1.0)

    # Evaluate the time derivative of the polynomial at all collocation points to get the coefficients of the continuity equation
    pder = numpy.polyder(p)
    for r in range(d+1):
        C[j,r] = pder(tau_root[r])

    # Evaluate the integral of the polynomial to get the coefficients of the quadrature function
    pint = numpy.polyint(p)
    B[j] = pint(1.0)

# Time horizon
T = 10.

# Declare model variables
x1 = MX.sym('x1')
x2 = MX.sym('x2')
x = vertcat([x1, x2])
u = MX.sym('u')

# Model equations
xdot = vertcat([(1-x2**2)*x1 - x2 + u, x1])

# Objective term
L = x1**2 + x2**2 + u**2;

# Continuous time dynamics
f = Function('f', [x, u], [xdot, L])

# Control discretization
N = 20 # number of control intervals
h = T/N

# Start with an empty NLP
w=[]
w0 = []
lbw = []
ubw = []
J = 0
g=[]
lbg = []
ubg = []

# "Lift" initial conditions
X0 = MX.sym('X0', 2)
w += [X0]
lbw += [0, 1]
ubw += [0, 1]
w0 += [0, 1]

# Formulate the NLP
Xk = MX([0, 1])
for k in range(N):
    # New NLP variable for the control
    Uk = MX.sym('U_' + str(k))
    w   += [Uk]
    lbw += [-1]
    ubw += [1]
    w0  += [0]

    # State at collocation points
    Xkj = []
    for j in range(d):
        Xkj += [MX.sym('X_'+str(k)+'_'+str(j), 2)];
        w += [Xkj[-1]]
        lbw += [-0.25, -inf]
        ubw += [inf,  inf]
        w0 += [0, 0]

    # Loop over collocation points
    Xk_end = D[0]*Xk
    for j in range(1,d+1):
       # Expression for the state derivative at the collocation point
       xp = C[0,j]*Xk
       for r in range(d):
           xp = xp + C[r+1,j]*Xkj[r];
      
       # Append collocation equations
       [fj, qj] = f([Xkj[j-1],Uk])
       g += [h*fj - xp]
       lbg += [0, 0]
       ubg += [0, 0]
       
       # Add contribution to the end state
       Xk_end = Xk_end + D[j]*Xkj[j-1];
  
       # Add contribution to quadrature function
       J = J + B[j]*qj*h

    # New NLP variable for state at end of interval
    Xk = MX.sym('X_' + str(k+1), 2)
    w   += [Xk]
    lbw += [-0.25, -inf]
    ubw += [  inf,  inf]
    w0  += [0, 0]

    # Add equality constraint
    g   += [Xk_end-Xk]
    lbg += [0, 0]
    ubg += [0, 0]

# Create an NLP solver
prob = {'f': J, 'x': vertcat(w), 'g': vertcat(g)}
solver = nlpsol('solver', 'ipopt', prob);

# Solve the NLP
arg = {'x0': w0, 'lbx': lbw, 'ubx': ubw, \
             'lbg': lbg, 'ubg': ubg}
sol = solver(arg)
w_opt = sol['x'].full().flatten()

# Plot the solution
x1_opt = w_opt[0::3+2*d]
x2_opt = w_opt[1::3+2*d]
u_opt = w_opt[2::3+2*d]

tgrid = [T/N*k for k in range(N+1)]
import matplotlib.pyplot as plt
plt.figure(1)
plt.clf()
plt.plot(tgrid, x1_opt, '--')
plt.plot(tgrid, x2_opt, '-')
plt.step(tgrid, vertcat((DM.nan(1), u_opt)), '-.')
plt.xlabel('t')
plt.legend(['x1','x2','u'])
plt.grid()
plt.show()
